theory LTS_Parser
imports 
  Xmlt2 
  "Proof-Checker.IA_Instance"
begin

hide_const (open) name

(* move? *)

context
  fixes atom_parser :: "('f,'a,'t) exp formula xmlt2"
begin

function formula_parser :: "('f,'a,'t) exp formula xmlt2" where
  "formula_parser x = (
  XMLdo ''disjunction'' {
    fs \<leftarrow>* formula_parser; xml_return (Disjunction fs)
  } XMLor XMLdo ''conjunction'' {
    fs \<leftarrow>* formula_parser; xml_return (Conjunction fs)
  } XMLor atom_parser) x" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

end

context
  fixes location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt2"
begin

definition transition_parser :: "('tr \<times> ('f,'v,'t,'l) transition_rule) xmlt2" where
  "transition_parser \<equiv> XMLdo ''transition'' {
    tr \<leftarrow> trans_parser;
    l \<leftarrow> xml_do ''source'' (xml_take location_parser xml_return);
    r \<leftarrow> xml_do ''target'' (xml_take location_parser xml_return);
    \<phi> \<leftarrow> xml_do ''formula'' (xml_take (formula_parser tatom_parser) xml_return);
    xml_return (tr, Transition l r \<phi>)
  }" 

definition lts_parser :: "tag \<Rightarrow> ('f,'v,'t,'l,'tr) lts_impl xmlt2" where
  "lts_parser tag \<equiv> XMLdo tag {
     i \<leftarrow> xml_do ''initial'' (xml_take_many 1 \<infinity> location_parser xml_return);
     t \<leftarrow>* transition_parser;
    xml_return (Lts_Impl i t [])
   }"

definition "safety_input_parser \<equiv>
  XMLdo ''ltsSafetyInput'' {
    lts \<leftarrow> lts_parser ''lts'';
    err \<leftarrow> xml_do ''error'' (xml_take_many 1 \<infinity> location_parser xml_return);
    xml_return (lts,err)
  }"

end

context
  fixes art_node_id_parser :: "tag \<Rightarrow> 'n xmlt2"
    and location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and hint_parser :: "'h :: default xmlt2"
begin

definition formula_pos_parser :: "nat xmlt2" where [code]:
 "formula_pos_parser =
  XMLdo ''conclusion'' {xml_return 0}
  XMLor XMLdo ''assertion'' {xml_return 2}
  XMLor XMLdo ''transition'' {xml_return 3}
  XMLor XMLdo ''targetAssertion'' {xml_return 4}"

function hints_parser :: "'h hint xmlt2" where
 "hints_parser xml = (
  XMLdo ''auto'' {
    xml_return default
  } XMLor XMLdo ''distribute'' {
    pos \<leftarrow> formula_pos_parser;
    hints \<leftarrow>* hints_parser;
    xml_return (hint.Distribute pos hints)
  } XMLor XMLdo ''erase'' {
    pos \<leftarrow> formula_pos_parser;
    hint \<leftarrow> hints_parser;
    xml_return (hint.Erase pos hint)
  } XMLor XMLdo ''lexWeak'' {
    hints \<leftarrow>* hints_parser;
    xml_return (hint.LexWeak hints)
  } XMLor XMLdo ''lexStrict'' {
    hints \<leftarrow>* hints_parser;
    xml_return (hint.LexStrict hints)
  } XMLor
    xml_change hint_parser (xml_return \<circ> hint.Base)
) xml" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

definition art_node_parser :: "(('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_node_impl \<times> 'n list) xmlt2" where
 "art_node_parser \<equiv> XMLdo ''node'' {
    init \<leftarrow>[False] (xml_do ''initial'' (xml_return True));
    nodeId \<leftarrow> art_node_id_parser ''nodeId'';
    invariant \<leftarrow> xml_do ''invariant'' (xml_take (formula_parser atom_parser) xml_return);
    location \<leftarrow> xml_do ''location'' (xml_take location_parser xml_return);
    edges \<leftarrow> XMLdo ''children'' {
        chs \<leftarrow>*
          XMLdo ''child'' {
            tr \<leftarrow> trans_parser;
            n \<leftarrow> art_node_id_parser ''nodeId'';
            h \<leftarrow>[default] (xml_do ''hints'' (xml_take hints_parser xml_return));
            xml_return (tr, n, h)
          };
        xml_return (Children_Edge chs)
      } XMLor XMLdo ''coverEdge'' {
          n \<leftarrow> art_node_id_parser ''nodeId'';
          h \<leftarrow>[default] xml_do ''hints'' (xml_take hints_parser xml_return);
          xml_return (Cover_Edge n h)
       };
    xml_return (Art_Node nodeId invariant location edges, if init then [nodeId] else [])
  }"

definition art_parser :: "(('f, 'v, 't, 'l, 'n, 'tr, 'h hint) art_impl) xmlt2" where
 "art_parser \<equiv> XMLdo ''impact'' {
    init \<leftarrow>[[]] xml_change (art_node_id_parser ''initial'') (\<lambda>x. xml_return [x]);
    pairs \<leftarrow> xml_do ''nodes'' (xml_take_many 0 \<infinity> art_node_parser xml_return);
    let nodes = map fst pairs;
    let inits = init @ concat (map snd pairs);
    xml_return \<lparr> art_impl.initial_nodes = inits, nodes = nodes \<rparr>
  }"

end

context
  fixes location_parser :: "'l xmlt2"
    and trans_parser :: "'tr xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and hint_parser :: "'h :: default xmlt2"
begin

type_synonym art_node_id = string

abbreviation art where "art \<equiv> art_parser xml_text location_parser trans_parser atom_parser hint_parser"

definition "invariant_proof_parser I \<equiv>
  xml_change art (\<lambda> prf. xml_return (Impact I prf))"

definition invariant_parser :: "('l \<times> ('f,'v,'t) exp formula) xmlt2" where
 "invariant_parser \<equiv> XMLdo ''invariant'' {
    l \<leftarrow> xml_do ''location'' (xml_take location_parser xml_return);
    \<phi> \<leftarrow> xml_do ''formula'' (xml_take (formula_parser atom_parser) xml_return);
    xml_return (l,\<phi>)
  }"

definition invariants_parser :: "('l \<times> ('f,'v,'t) exp formula)list xmlt2" where
 "invariants_parser \<equiv> xml_do ''invariants'' (xml_take_many 0 \<infinity> invariant_parser xml_return)"

definition safety_proof_parser :: "('f,'v,'t,'l,string,'tr,'h) safety_proof xmlt2" where
 "safety_proof_parser \<equiv> XMLdo ''safetyViaInvariants'' {
    I \<leftarrow> invariants_parser;
    p \<leftarrow> invariant_proof_parser I;
    xml_return (safety_proof.Invariant_Assertion p safety_proof.Trivial)
  }"

end


fun cut_points_to_transitions :: "_ \<Rightarrow> _ \<Rightarrow> ('f,'v,'t,'l sharp,'tr sharp) transitions_impl" where
  "cut_points_to_transitions ts [] = ts"
| "cut_points_to_transitions ts ((l,tr,\<phi>)#cps) =
   cut_points_to_transitions ((Flat tr, Transition (Flat l) (Sharp l) \<phi>)#ts) cps"

context
  fixes location_id_parser :: "tag \<Rightarrow> 'l xmlt2"
    and trans_id_parser :: "tag \<Rightarrow> 'tr xmlt2"
    and exp_parser :: "('f,'v,'t) exp xmlt2"
    and atom_parser :: "('f,'v,'t) exp formula xmlt2"
    and tatom_parser :: "('f,'v trans_var,'t) exp formula xmlt2"
    and variable_parser :: "'v xmlt2"
    and hint_parser :: "'h :: default xmlt2"
    and type_parser :: "'t xmlt2"
    and dom_type :: "'t"
begin

definition "location_parser \<equiv> location_id_parser ''locationId''"

definition sharp_location_parser :: "'l sharp xmlt2" where
 "sharp_location_parser \<equiv>
    (xml_change (location_id_parser ''locationDuplicate'') (xml_return \<circ> Sharp))
    XMLor (xml_change (location_id_parser ''locationId'') (xml_return \<circ> Flat))"

abbreviation "tformula \<equiv> formula_parser tatom_parser"

context
  fixes trans_parser :: "'trr xmlt2"
begin

abbreviation "lts \<equiv> lts_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "transition \<equiv> transition_parser sharp_location_parser trans_parser tatom_parser"

abbreviation "invariant_proof \<equiv> invariant_proof_parser sharp_location_parser trans_parser atom_parser hint_parser"

abbreviation "invariants \<equiv> invariants_parser sharp_location_parser atom_parser"

function cooperation_proof_parser where "cooperation_proof_parser xs = (
  xml_do ''trivial'' (xml_return cooperation_proof.Trivial)
  XMLor XMLdo ''newInvariants'' {
    i \<leftarrow> invariants;
    p \<leftarrow> invariant_proof i;
    cp \<leftarrow> cooperation_proof_parser;
    xml_return (Invariants_Update p cp)
  } XMLor XMLdo ''transitionRemoval'' {
    rs \<leftarrow> XMLdo ''rankingFunctions'' {
      pairs \<leftarrow>* XMLdo ''rankingFunction'' {
        l \<leftarrow> xml_do ''location'' (xml_take sharp_location_parser xml_return);
        es \<leftarrow> xml_do ''expression'' (xml_take_many 1 \<infinity> exp_parser xml_return);
        xml_return (l,es)
      };
      xml_return pairs};
    bounds \<leftarrow> xml_do ''bound'' (xml_take_many 0 \<infinity> exp_parser xml_return);
    removed \<leftarrow> xml_do ''remove'' (xml_take_many 0 \<infinity> trans_parser xml_return);
    hinter \<leftarrow>[\<lambda>x. default] XMLdo ''hints'' {
      pairs \<leftarrow>* XMLdo ''hint'' {
        tr \<leftarrow> trans_parser;
        hint \<leftarrow>[default] hints_parser hint_parser;
        xml_return (tr, hint)
      };
      xml_return (map_of_default default pairs)};
    inner \<leftarrow> cooperation_proof_parser;
    let rf = map_of_default [] rs;
    xml_return (Transition_Removal (Transition_removal_info rf removed dom_type bounds hinter) inner)
  } XMLor XMLdo ''locationAddition'' {
    tr \<leftarrow> transition;
    prof \<leftarrow> cooperation_proof_parser;
    let (tr_id,tau) = tr;
    xml_return (Location_Addition (Location_Addition_Info (source tau) (target tau) tr_id tau) prof)
  } XMLor XMLdo ''freshVariableAddition'' {
    x1 \<leftarrow> variable_parser;
    x2 \<leftarrow> type_parser;
    x3 \<leftarrow> XMLdo ''additionalFormulas'' {
      pairs \<leftarrow>* XMLdo ''additionalFormula'' {
         tr \<leftarrow> trans_parser;
         \<phi> \<leftarrow> tformula;
         xml_return (tr,\<phi>)
      };
      xml_return pairs
    };
    prof \<leftarrow> cooperation_proof_parser;
    xml_return (Fresh_Variable_Addition (Fresh_Variable_Addition_Info x1 x2 x3) prof)
  } XMLor XMLdo ''cutTransitionSplit'' {
    ps \<leftarrow>* (XMLdo ''cutTransitionsWithProof'' {
      cuts \<leftarrow> xml_do ''cutTransitions'' (xml_take_many 0 \<infinity> trans_parser xml_return);
      prof \<leftarrow> cooperation_proof_parser;
      xml_return (cuts,prof)});
    xml_return (Cut_Transition_Split ps)
  } XMLor XMLdo ''sccDecomposition'' {
    sccs \<leftarrow>* XMLdo ''sccWithProof'' {
      scc :: 'l sharp list \<leftarrow> xml_do ''scc'' (xml_take_many 1 \<infinity> sharp_location_parser xml_return);
      prof \<leftarrow> cooperation_proof_parser;
      xml_return (scc,prof)};
    xml_return (Scc_Decomp sccs)
  }) xs" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

end

definition "trans_id \<equiv> trans_id_parser ''transitionId''"


definition "sharp_trans_id \<equiv>
  xml_change (trans_id_parser ''transitionDuplicate'') (xml_return \<circ> Sharp)
  XMLor xml_change (trans_id_parser ''transitionId'') (xml_return \<circ> Flat)"

definition "cutPoints_parser \<equiv> XMLdo ''cutPoints'' {
  tuples \<leftarrow>* XMLdo ''cutPoint'' {
    l \<leftarrow> location_id_parser ''locationId'';
    tr \<leftarrow> trans_id_parser ''skipId'';
    \<phi> \<leftarrow> XMLdo ''skipFormula'' {
      ret \<leftarrow> tformula;
      xml_return ret
    };
    xml_return (l,tr,\<phi>)
  };
  xml_return (cut_points_to_transitions [] tuples)
}"

function termination_proof_parser where "termination_proof_parser xml = (
  XMLdo ''trivial'' {
    xml_return termination_proof.Trivial
  } XMLor XMLdo ''newInvariants'' {
    i \<leftarrow> invariants_parser location_parser atom_parser;
    p \<leftarrow> invariant_proof_parser location_parser trans_id atom_parser hint_parser i;
    cp \<leftarrow> termination_proof_parser;
    xml_return (Invariants_Update_LTS p cp)
  } XMLor XMLdo ''switchToCooperationTermination'' {
    cp \<leftarrow> cutPoints_parser;
    p \<leftarrow> cooperation_proof_parser sharp_trans_id;
    xml_return (termination_proof.Via_Cooperation [(cp,p)])
  }
) xml" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

end


type_synonym location_id = string
type_synonym transition_id = string
type_synonym variable_id = string


definition variable_parser :: "variable_id xmlt2" where
  "variable_parser = xml_text ''variableId''"

definition trans_var_parser :: "variable_id trans_var xmlt2" where
  "trans_var_parser =
    XMLdo ''post'' {v \<leftarrow> variable_parser; xml_return (Post v)}
    XMLor xml_change (xml_text ''aux'') (xml_return \<circ> Intermediate)
    XMLor xml_change variable_parser (xml_return \<circ> Pre)"

context IA_locale begin

definition "constant_parser \<equiv> xml_int ''constant''"

function exp_parser' :: "'a xmlt2 \<Rightarrow> 'a IA_exp xmlt2" where
  "exp_parser' v xml = (
  XMLdo ''product'' {
    exps \<leftarrow>* exp_parser' v;
    xml_return (Fun (ProdF (length exps)) exps)
  } XMLor XMLdo ''sum'' {
    exps \<leftarrow>* exp_parser' v;
    xml_return (Fun (SumF (length exps)) exps)
  } XMLor xml_change (xml_int ''constant'') (\<lambda>i. xml_return (Fun (ConstF i) []))
  XMLor xml_change v (\<lambda> x. xml_return (Var (x, IntT)))
  ) xml"
by (pat_completeness, auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)


definition bexp_parser' :: "'a :: show xmlt2 \<Rightarrow> 'a IA_exp xmlt2" where
 "bexp_parser' v \<equiv>
  XMLdo ''leq'' {
    l \<leftarrow> exp_parser' v;
    r \<leftarrow> exp_parser' v;
    xml_return (Fun LeF [l, r])
  } XMLor XMLdo ''less'' {
    l \<leftarrow> exp_parser' v;
    r \<leftarrow> exp_parser' v;
    xml_return (Fun LessF [l,r])
  } XMLor XMLdo ''eq'' {
    l \<leftarrow> exp_parser' v;
    r \<leftarrow> exp_parser' v;
    xml_return (Fun EqF [l,r])
  }"

definition "exp_parser \<equiv> exp_parser' variable_parser" 
definition "atom_parser \<equiv> xml_change (bexp_parser' variable_parser) (xml_return \<circ> Atom)"
definition "tatom_parser \<equiv> xml_change (bexp_parser' trans_var_parser) (xml_return \<circ> Atom)"
definition "atom_parser' = xml_change (bexp_parser' variable_parser) (xml_return \<circ> Atom)"
definition "tatom_parser' = xml_change (bexp_parser' trans_var_parser) (xml_return \<circ> Atom)"

definition hint_parser :: "hints xmlt2" where
  "hint_parser \<equiv> XMLdo ''linearCombination'' {
      xs \<leftarrow>* xml_int ''constant'';
      xml_return (Hints xs)
  }"

definition type_parser :: "ty xmlt2" where
  "type_parser = xml_do ''int'' (xml_return IntT) XMLor xml_do ''bool'' (xml_return BoolT)"


definition "lts_input_parser \<equiv>
  lts_parser
    (location_parser xml_text)
    (trans_id xml_text) tatom_parser' ''lts''"

definition "lts_safety_input_parser \<equiv>
  safety_input_parser
    (location_parser xml_text)
    (trans_id xml_text)
    tatom_parser'"

definition "lts_termination_proof_parser \<equiv>
  termination_proof_parser
    xml_text
    xml_text
    exp_parser atom_parser tatom_parser variable_parser hint_parser type_parser IntT"

definition "lts_safety_proof_parser \<equiv>
  safety_proof_parser (location_parser xml_text) (trans_id xml_text) atom_parser hint_parser"

end

end
