(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Code_Setup_Haskell
imports
  Util
begin

subsection \<open>Code-Generator Setup\<close>

text \<open>The order of the final two arguments of @{const foldr} in
Isabelle is different from Haskell's @{text foldr}\<close>
definition flipped_foldr where "flipped_foldr f \<equiv> flip (foldr f)"
text \<open>Separate infix use and prefix use of infix operators\<close>
definition comp_infix where "comp_infix x y \<equiv> x \<circ> y"
definition Cons_infix where "Cons_infix x y \<equiv> x # y"

lemma foldr_flipped_foldr[code_unfold]:
  "foldr f xs a = flipped_foldr f a xs"
  by (simp add: flipped_foldr_def)

lemma flip_foldr_flipped_foldr[code_unfold]:
  "flip (foldr f) = flipped_foldr f"
  by (simp add: flipped_foldr_def)

lemma o_comp_infix[code_unfold]: "x \<circ> y = comp_infix x y"
  by (simp add: comp_infix_def)

code_const const (Haskell "const")
code_const flip (Haskell "flip")
code_const flipped_foldr (Haskell "foldr")
code_const "op \<circ>" (Haskell "(.)")
code_const comp_infix (Haskell infixr 9 ".")
code_const "op #" (Haskell "(:)")
code_const Cons_infix (Haskell infixr 5 ":")
(*
(*Problem: Haskell's "length" is of type "[a] -> Int", whereas
  Isabelle's "length" is of type "'a list => nat" (thus having
  arbitrary precision). Hence, no direct mapping is possible/desirable.*)
code_const length (Haskell "length")
*)

text \<open>NOTE: the definition of comp_infix together with o_comp_infix leads to
nontermination for all target languages that do not have a proper code_const setup
for comp_infix\<close>

end
