(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2016)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2012)
License: LGPL (see file COPYING.LESSER)
*)
theory Trs_Impl
  imports
    Trs
    Term_Impl
    Matching
    Abstract_Rewriting_Impl
    "../Auxiliaries/Option_Util"
    "../Framework/Map_Choice"
    "Transitive-Closure.RBT_Map_Set_Extension"
    "Renaming_Interpretations"
    "../Normalization_Equivalence/Encompassment"
begin

type_synonym ('f, 'v) rules = "('f, 'v) rule list"

fun rewrite :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "rewrite R s = concat (map (\<lambda> (l, r) . concat (map (\<lambda> p.
    (case match (s |_ p) l of
      None \<Rightarrow> []
    | Some \<sigma> \<Rightarrow> [replace_at s p (r \<cdot> \<sigma>)])) (poss_list s))) R)"

lemma rewrite_complete_main:
  fixes l r :: "('f, 'v) term"
  assumes "(s, t) \<in> rstep_r_p_s (set R) (l, r) p \<sigma>" 
  shows "\<exists>u \<in> set (rewrite R s). vars_term l \<supseteq> vars_term r \<longrightarrow> u = t"
proof -
  let ?C = "ctxt_of_pos_term p s"
  from assms have lr: "(l,r) \<in> set R" 
    and p: "p \<in> poss s" 
    and s: "?C\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t: "?C\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from s p have "?C\<langle>l \<cdot> \<sigma>\<rangle> = ?C\<langle>s|_p\<rangle>" by (simp add: ctxt_supt_id)
  hence "l \<cdot> \<sigma> = s|_p" by auto
  from this match_complete' obtain \<delta> where 
    match: "match (s|_p) l = Some \<delta>" and 
      vars: "\<forall> x \<in> vars_term l. \<sigma> x = \<delta> x" by blast
  have l: "l \<cdot> \<sigma> = l \<cdot> \<delta>"
    by (rule term_subst_eq, insert vars, auto)
  obtain u where u: "u = ?C\<langle>r \<cdot> \<delta>\<rangle>" by auto
  {
    assume sub: "vars_term r \<subseteq> vars_term l"
    have "r \<cdot> \<sigma> = r \<cdot> \<delta>"
      by (rule term_subst_eq, insert sub vars, auto)
    with t have "u = t" unfolding u by simp
  } note ut = this
  have nearlyDone: "(l,r) \<in> set R \<and> p \<in> poss s \<and> u \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ ?C\<langle>r\<cdot>\<sigma>\<rangle>])" using lr p match unfolding u by auto 
  hence "u \<in> set ((\<lambda>(l, r). concat  (map 
     (\<lambda>p. case match (s|_p) l  of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>])
                          (poss_list s)))  (l,r))" (is "_ \<in> (?P (l,r))")
    by auto
  with lr have "\<exists> lr \<in> set R. u \<in> (?P lr)" by blast
  hence "u \<in> set (rewrite R s)" by auto
  thus ?thesis using ut by (auto simp del:rewrite.simps)
qed

lemma rewrite_complete:
  fixes l r::"('f,'v)term"
  assumes "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" 
  and "vars_term l \<supseteq> vars_term r"
  shows "t \<in> set (rewrite R s)"
  using rewrite_complete_main[OF assms(1)] using assms(2) by auto

lemma rewrite_sound:
  assumes "t \<in> set (rewrite R s)"
  shows "(s, t) \<in> rstep (set R)"
proof -
  from assms 
  have "\<exists> l r p.  (l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])" 
    by auto
  from this obtain l r p where one:
    "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])"
    by auto
  from this obtain \<sigma> where two: "match (s|_p) l = Some \<sigma> \<and> t \<in> {(ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>}"
    by (cases "match (s|_p) l", auto)
  hence match: "l \<cdot> \<sigma> = s|_p" using match_sound by auto 
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  thus "(s,t) \<in> rstep (set R)" unfolding rstep_iff_rstep_r_p_s by blast
qed

declare rewrite.simps[simp del]

lemma rewrite_iff_rstep:
  assumes "\<And> l r. (l,r) \<in> set R \<Longrightarrow> vars_term l \<supseteq> vars_term r"
  shows "set (rewrite R s) = {t. (s,t) \<in> rstep (set R)}" (is "?l = ?r")
proof -
  {
    fix t
    assume "t \<in> ?l"
    from rewrite_sound[OF this] have "t \<in> ?r" by auto
  }
  moreover
  {
    fix t
    assume "t \<in> ?r"
    hence "(s,t) \<in> rstep (set R)" by auto
    from this[unfolded rstep_iff_rstep_r_p_s] obtain l r p \<sigma> where 
      st: "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" by auto
    hence "(l,r) \<in> set R" unfolding rstep_r_p_s_def Let_def by auto
    from rewrite_complete[OF st assms[OF this]] have "t \<in> ?l" .
  }
  ultimately show ?thesis by auto
qed

definition first_rewrite :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term option"
  where "first_rewrite R s \<equiv> case rewrite R s of Nil \<Rightarrow> None | Cons t _ \<Rightarrow> Some t"

definition in_rstep_impl where 
  "in_rstep_impl s t R \<equiv> \<exists> p \<in> set (poss_list s). p \<in> poss t \<and> ctxt_of_pos_term p s = ctxt_of_pos_term p t \<and> 
     (let sp = s |_ p; tp = t |_ p in \<exists> lr \<in> R. match_list Var [(fst lr,sp), (snd lr,tp)] \<noteq> None)"

lemma in_rstep_code[code_unfold]: "(st \<in> rstep R) = (in_rstep_impl (fst st) (snd st) R)"
proof (cases st)
  case (Pair s t)
  show ?thesis unfolding Pair in_rstep_impl_def Let_def fst_conv snd_conv poss_list_sound
  proof -
    show "((s,t) \<in> rstep R) = 
      (\<exists>p\<in>poss s. p \<in> poss t \<and> ctxt_of_pos_term p s = ctxt_of_pos_term p t \<and> (\<exists>lr\<in>R. match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] \<noteq> None))"
      (is "?l = ?r")
    proof
      assume ?l
      from this[unfolded rstep_iff_rstep_r_p_s] obtain l r p \<sigma>
      where "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>" by auto
      from this[unfolded rstep_r_p_s_def Let_def fst_conv snd_conv]
      have ps: "p \<in> poss s" and lr: "(l, r) \<in> R" and s: "(ctxt_of_pos_term p s)\<langle>l \<cdot> \<sigma>\<rangle> = s" and t: "(ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle> = t" by auto
      have pt: "p \<in> poss t" using ps s t 
        by (metis hole_pos_ctxt_of_pos_term hole_pos_poss)
      from ps pt have id: "ctxt_of_pos_term p s = ctxt_of_pos_term p t" unfolding t[symmetric]
        by (metis hole_pos_ctxt_of_pos_term hole_pos_id_ctxt)
      from s ps have sp: "s |_ p = l \<cdot> \<sigma>" by (rule subt_at_ctxt_of_pos_term)
      from t ps have tp: "t |_ p = r \<cdot> \<sigma>" by (metis hole_pos_ctxt_of_pos_term subt_at_hole_pos)
      let ?P = "[(l, s |_ p), (r, t |_ p)]"
      have "\<And> sa ta. (sa, ta) \<in> set ?P \<Longrightarrow> sa \<cdot> \<sigma> = ta" unfolding sp tp by auto
      from match_list_complete'[of ?P \<sigma> Var, OF this] have match: "match_list Var ?P \<noteq> None" by auto
      show ?r
        by (intro bexI conjI, rule pt, insert lr id match ps, auto)
    next
      assume ?r
      then obtain p lr where 
        ps: "p \<in> poss s" and
        pt: "p \<in> poss t" and
        id: "ctxt_of_pos_term p s = ctxt_of_pos_term p t" (is "?Cs = ?Ct") and
        lr: "lr\<in>R" and
        match: "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] \<noteq> None" 
        by blast
      obtain l r where l_r: "lr = (l,r)" by force
      from lr l_r have lr: "(l,r) \<in> R" by simp
      from match l_r obtain \<sigma> where "match_list Var [(l, s|_p), (r, t|_p)] = Some \<sigma>" by force
      from match_list_sound[OF this, unfolded matchers_def] have sp: "l \<cdot> \<sigma> = s |_ p" and tp: "r \<cdot> \<sigma> = t |_ p" by auto
      from ps have s: "s = ?Cs \<langle> l \<cdot> \<sigma> \<rangle>" unfolding sp by (metis ctxt_supt_id)
      from pt have t: "t = ?Cs \<langle> r \<cdot> \<sigma> \<rangle>" unfolding tp id by (metis ctxt_supt_id)
      show ?l by (rule rstepI[OF lr s t])
    qed
  qed
qed

definition compute_rstep_NF :: "('f,'v)rules \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term option"
  where "compute_rstep_NF R s \<equiv> compute_NF (first_rewrite R) s"

lemma compute_rstep_NF_sound:
  assumes res: "compute_rstep_NF R s = Some t"
  shows "(s, t) \<in> (rstep (set R))^*" using res[unfolded compute_rstep_NF_def]
proof (rule compute_NF_sound)
  fix s t
  assume "first_rewrite R s = Some t"
  from this[unfolded first_rewrite_def] obtain ts where "rewrite R s = t # ts"
    by (cases "rewrite R s", auto)
  hence t: "t \<in> set (rewrite R s)" by simp
  from rewrite_sound[OF this] show "(s,t) \<in> rstep (set R)" .
qed

lemma compute_rstep_NF_complete: assumes res: "compute_rstep_NF R s = Some t"
  shows "t \<in> NF (rstep (set R))" using res[unfolded compute_rstep_NF_def]
proof (rule compute_NF_complete)
  fix s
  assume "first_rewrite R s = None"
  from this[unfolded first_rewrite_def] have empty: "rewrite R s = []" 
    by (cases "rewrite R s", auto)
  show "s \<in> NF (rstep (set R))"
  proof(rule, rule)
    fix t
    assume "(s,t) \<in> rstep (set R)"
    from this[unfolded rstep_iff_rstep_r_p_s]
    obtain l r p \<sigma> where step: "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" by auto
    from rewrite_complete_main[OF step] arg_cong[OF empty, of set] show False 
      by auto
  qed
qed

lemma compute_rstep_NF_SN: assumes SN: "SN (rstep (set R))"
  shows "\<exists> t. compute_rstep_NF R s = Some t"
proof -
  have "\<exists> t. compute_NF (first_rewrite R) s = Some t"
  proof (rule compute_NF_SN[OF SN])
    fix s t
    assume "first_rewrite R s = Some t"
    from this[unfolded first_rewrite_def] have
      rewrite: "t \<in> set (rewrite R s)" by (auto split: list.splits)
    from rewrite_sound[OF this]
    show "(s,t) \<in> rstep (set R)" .
  qed  
  thus ?thesis unfolding compute_rstep_NF_def .
qed

definition
  check_join_NF ::
    "('f :: show, 'v :: show) rules \<Rightarrow> 
     ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_join_NF R s t \<equiv> case (compute_rstep_NF R s, compute_rstep_NF R t) of
      (Some s', Some t') \<Rightarrow> 
     check (s' = t') (
    ''the normal form '' +#+ shows s' +@+ shows '' of '' +@+ shows s +@+ shows '' differs from '' +@+ shows_nl +@+ shows ''the normal form '' +@+ shows t' +@+ shows '' of '' +@+ shows t)
       |  _ \<Rightarrow> error (shows ''strange error in normal form computation'')"

lemma check_join_NF_sound:
  assumes ok: "isOK (check_join_NF R s t)"
  shows "(s, t) \<in> join (rstep (set R))"
proof -
  note ok = ok[unfolded check_join_NF_def]
  from ok obtain s' where s: "compute_rstep_NF R s = Some s'" by force
  note ok = ok[unfolded s]
  from ok obtain t' where t: "compute_rstep_NF R t = Some t'" by force
  from ok[unfolded t] have id: "s' = t'" by simp
  note seq = compute_rstep_NF_sound
  from seq[OF s] seq[OF t]
  show ?thesis unfolding id by auto
qed

fun reachable_terms :: 
  "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> nat \<Rightarrow> ('f, 'v) term list"
where 
  "reachable_terms R s 0 = [s]"
| "reachable_terms R s (Suc n) = (
  let ts = (reachable_terms R s n) in
    remdups (ts@(concat (map (\<lambda> t. rewrite R t) ts)))
  )"

lemma reachable_terms_nat:
assumes "t \<in> set (reachable_terms R s i)"
shows "\<exists> j. j \<le> i \<and> (s,t) \<in> (rstep (set R))^^j"
using assms
proof (induct i arbitrary: t)
  case 0 
  thus ?case by auto
next
  case (Suc i) 
  let ?R = "\<lambda> j. (rstep (set R))^^j"
  from Suc(2)
  have "t \<in> set (reachable_terms R s i)
    \<or> (\<exists> u \<in> set (reachable_terms R s i). t \<in> set (rewrite R u))" by (simp add: Let_def)
  thus ?case
  proof
    assume "t \<in> set (reachable_terms R s i)"
    from Suc(1)[OF this] obtain j where "j \<le> i" and "(s,t) \<in> ?R j" by auto 
    thus ?thesis by (intro exI[of _ j], auto)
  next
    assume "\<exists> u \<in> set (reachable_terms R s i). t \<in> set (rewrite R u)"
    then obtain u where u: "u \<in> set (reachable_terms R s i)"
      and 1: "t \<in> set (rewrite R u)" by auto
    from rewrite_sound[OF 1] have ut: "(u,t) \<in> rstep (set R)" by auto
    from Suc(1)[OF u] obtain j where j: "j \<le> i" and su: "(s,u) \<in> ?R j" by auto
    from su ut have "(s,t) \<in> ?R (Suc j)" by auto
    with j show ?thesis by (intro exI[of _ "Suc j"], auto)
  qed
qed

lemma reachable_terms:
assumes "t \<in> set (reachable_terms R s i)"
shows "(s,t) \<in> (rstep (set R))^*"
  using reachable_terms_nat[OF assms] by (metis relpow_imp_rtrancl)

lemma reachable_terms_one:
  assumes "t \<in> set (reachable_terms R s (Suc 0))"
  shows "(s,t) \<in> (rstep (set R))^="
proof -
  from reachable_terms_nat[OF assms] obtain j where "j \<le> 1"
  and "(s,t) \<in> (rstep (set R))^^j" by auto
  thus ?thesis by (cases j, auto)
qed


function iterative_join_search_main ::
  "('f,'v) rules \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool"
where
  "iterative_join_search_main R s t i n = (if i \<le> n then
  (((list_inter (reachable_terms R s i) (reachable_terms R t i)) \<noteq> []) \<or> (iterative_join_search_main R s t (Suc i) n)) else False)"
by pat_completeness auto

termination by (relation "measure ( \<lambda> (R,s,t,i,n). Suc n - i)") auto

lemma iterative_join_search_main: 
  "iterative_join_search_main R s t i n \<Longrightarrow> (s,t) \<in> join (rstep (set R))"
proof (induction rule: iterative_join_search_main.induct)
  case (1 R s t i n)
  from 1(2) have i_n: "i \<le> n" by (simp split: if_splits)
  note IH = 1(1)[OF i_n]
  let ?I = "list_inter (reachable_terms R s i) (reachable_terms R t i)"
  from 1(2) i_n have "?I \<noteq> [] \<or> iterative_join_search_main R s t (Suc i) n" by auto
  thus ?case
  proof
    assume a: "?I \<noteq> []"
    then obtain u us where u: "?I = u # us" by (cases ?I, auto)
    hence d: "u \<in> set ?I" by auto
    from this[simplified] reachable_terms[of u _ _ i] have c: "(s, u) \<in> (rstep (set R))\<^sup>*" by auto
    from d[simplified] reachable_terms[of u _ _ i] have e: "(t,u) \<in> (rstep (set R))^*" by auto
    from c e show ?thesis by auto
  next
    assume b: "iterative_join_search_main R s t (Suc i) n"
    from  IH[OF this] show ?thesis .
  qed
qed

definition iterative_join_search where 
  "iterative_join_search R s t n \<equiv> iterative_join_search_main R s t 0 n"

lemma iterative_join_search: "iterative_join_search R s t n \<Longrightarrow> (s,t) \<in> join (rstep (set R))"
  by (rule iterative_join_search_main, unfold iterative_join_search_def)

definition
  check_join_BFS_limit ::
    "nat \<Rightarrow> ('f :: show, 'v :: show) rules \<Rightarrow> 
     ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_join_BFS_limit n R s t \<equiv> check (iterative_join_search R s t n) 
    (shows ''could not find a joining sequence of length at most '' +@+ 
     shows n +@+ shows '' for the terms '' +@+ shows s +@+ shows '' and '' +@+ shows t +@+ shows_nl)"

lemma check_join_BFS_limit_sound:
  assumes ok: "isOK (check_join_BFS_limit n R s t)"
  shows "(s, t) \<in> join (rstep (set R))"
  by (rule iterative_join_search, insert ok[unfolded check_join_BFS_limit_def], simp)

definition map_funs_rules :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('g, 'v) rules" where
  "map_funs_rules fg R = map (map_funs_rule fg) R"

lemma map_funs_rules_sound[simp]:
  "set (map_funs_rules fg R) = map_funs_trs fg (set R)"
  unfolding map_funs_rules_def map_funs_trs.simps by simp


subsection \<open>Output\<close>

fun shows_rule' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> ('f, 'v) rule \<Rightarrow> shows"
where
  "shows_rule' fun var arr (l, r) =
    shows_term' fun var l \<circ> shows arr \<circ> shows_term' fun var r"

abbreviation "shows_rule \<equiv> shows_rule' shows shows '' -> ''"
abbreviation "shows_weak_rule \<equiv> shows_rule' shows shows '' ->= ''"

definition
  shows_rules' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows"
where
  "shows_rules' fun var arr trs =
    shows_list_gen (shows_rule' fun var arr) [] [] [CHR ''\<newline>''] [] trs \<circ> shows_nl"

abbreviation "shows_rules \<equiv> shows_rules' shows shows '' -> ''"
abbreviation "shows_weak_rules \<equiv> shows_rules' shows shows '' ->= ''"

definition
  shows_trs' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> string \<Rightarrow> string \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows"
where
  "shows_trs' fun var name arr R = shows name \<circ> shows_nl \<circ> shows_nl \<circ> shows_rules' fun var arr R"

abbreviation "shows_trs \<equiv> shows_trs' shows shows ''rewrite system:'' '' -> ''"

definition add_vars_rule :: "('f, 'v) rule \<Rightarrow> 'v list \<Rightarrow> 'v list"
where
  "add_vars_rule r xs = add_vars_term (fst r) (add_vars_term (snd r) xs)"

definition add_funs_rule :: "('f, 'v) rule \<Rightarrow> 'f list \<Rightarrow> 'f list"
where
  "add_funs_rule r fs = add_funs_term (fst r) (add_funs_term (snd r) fs)"

definition add_funas_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "add_funas_rule r fs = add_funas_term (fst r) (add_funas_term (snd r) fs)"

definition add_roots_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "add_roots_rule r fs = root_list (fst r) @ root_list (snd r) @ fs"

definition add_funas_args_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "add_funas_args_rule r fs = add_funas_args_term (fst r) (add_funas_args_term (snd r) fs)"

lemma add_vars_rule_vars_rule_list_conv [simp]:
  "add_vars_rule r xs = vars_rule_list r @ xs"
  by (simp add: add_vars_rule_def vars_rule_list_def)

lemma add_funs_rule_funs_rule_list_conv [simp]:
  "add_funs_rule r fs = funs_rule_list r @ fs"
  by (simp add: add_funs_rule_def funs_rule_list_def)

lemma add_funas_rule_funas_rule_list_conv [simp]:
  "add_funas_rule r fs = funas_rule_list r @ fs"
  by (simp add: add_funas_rule_def funas_rule_list_def)

lemma add_roots_rule_roots_rule_list_conv [simp]:
  "add_roots_rule r fs = roots_rule_list r @ fs"
  by (simp add: add_roots_rule_def roots_rule_list_def)

lemma add_funas_args_rule_funas_args_rule_list_conv [simp]:
  "add_funas_args_rule r fs = funas_args_rule_list r @ fs"
  by (simp add: add_funas_args_rule_def funas_args_rule_list_def)

definition insert_vars_rule :: "('f, 'v) rule \<Rightarrow> 'v list \<Rightarrow> 'v list"
where
  "insert_vars_rule r xs = insert_vars_term (fst r) (insert_vars_term (snd r) xs)"

definition insert_funs_rule :: "('f, 'v) rule \<Rightarrow> 'f list \<Rightarrow> 'f list"
where
  "insert_funs_rule r fs = insert_funs_term (fst r) (insert_funs_term (snd r) fs)"

definition insert_funas_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_funas_rule r fs = insert_funas_term (fst r) (insert_funas_term (snd r) fs)"

definition insert_roots_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_roots_rule r fs =
    foldr List.insert (option_to_list (root (fst r)) @ option_to_list (root (snd r))) fs"

definition insert_funas_args_rule :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_funas_args_rule r fs = insert_funas_args_term (fst r) (insert_funas_args_term (snd r) fs)"

lemma set_insert_vars_rule [simp]:
  "set (insert_vars_rule r xs) = vars_term (fst r) \<union> vars_term (snd r) \<union> set xs"
  by (simp add: insert_vars_rule_def ac_simps)

lemma set_insert_funs_rule [simp]:
  "set (insert_funs_rule r xs) = funs_term (fst r) \<union> funs_term (snd r) \<union> set xs"
  by (simp add: insert_funs_rule_def ac_simps)

lemma set_insert_funas_rule [simp]:
  "set (insert_funas_rule r xs) = funas_term (fst r) \<union> funas_term (snd r) \<union> set xs"
  by (simp add: insert_funas_rule_def ac_simps)

lemma set_insert_roots_rule [simp]:
  "set (insert_roots_rule r xs) = root_set (fst r) \<union> root_set (snd r) \<union> set xs"
  by (cases "fst r" "snd r" rule: term.exhaust [case_product term.exhaust])
     (auto simp: insert_roots_rule_def ac_simps)

lemma set_insert_funas_args_rule [simp]:
  "set (insert_funas_args_rule r xs) = funas_args_term (fst r) \<union> funas_args_term (snd r) \<union> set xs"
  by (simp add: insert_funas_args_rule_def ac_simps funas_args_term_def)

abbreviation "vars_rule_impl r \<equiv> insert_vars_rule r []"
abbreviation "funs_rule_impl r \<equiv> insert_funs_rule r []"
abbreviation "funas_rule_impl r \<equiv> insert_funas_rule r []"
abbreviation "roots_rule_impl r \<equiv> insert_roots_rule r []"
abbreviation "funas_args_rule_impl r \<equiv> insert_funas_args_rule r []"

lemma set_vars_rule_impl:
  "set (vars_rule_impl r) = vars_rule r"
  by (simp add: vars_rule_def)

lemma [code]:
  "vars_rule_list r = add_vars_rule r []"
  "funs_rule_list r = add_funs_rule r []"
  "funas_rule_list r = add_funas_rule r []"
  "roots_rule_list r = add_roots_rule r []"
  "funas_args_rule_list r = add_funas_args_rule r []"
  by (simp_all add: vars_rule_list_def funs_rule_list_def funas_rule_list_def
                    roots_rule_list_def funas_args_rule_list_def)

lemma [code]:
  "vars_trs_list trs = foldr add_vars_rule trs []"
  "funs_trs_list trs = foldr add_funs_rule trs []"
  "funas_trs_list trs = foldr add_funas_rule trs []"
  "funas_args_trs_list trs = foldr add_funas_args_rule trs []"
  by (induct_tac [!] trs)
     (simp_all add: vars_trs_list_def funs_trs_list_def funas_trs_list_def
                    roots_trs_list_def funas_args_trs_list_def)

definition insert_vars_trs :: "('f, 'v) rule list \<Rightarrow> 'v list \<Rightarrow> 'v list"
where
  "insert_vars_trs trs = foldr insert_vars_rule trs"

definition insert_funs_trs :: "('f, 'v) rule list \<Rightarrow> 'f list \<Rightarrow> 'f list"
where
  "insert_funs_trs trs = foldr insert_funs_rule trs"

definition insert_funas_trs :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_funas_trs trs = foldr insert_funas_rule trs"

definition insert_roots_trs :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_roots_trs trs = foldr insert_roots_rule trs"

definition insert_funas_args_trs :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f \<times> nat) list"
where
  "insert_funas_args_trs trs = foldr insert_funas_args_rule trs"

lemma set_insert_vars_trs [simp]:
  "set (insert_vars_trs trs xs) = (\<Union>r \<in> set trs. vars_rule r) \<union> set xs"
  by (induct trs arbitrary: xs) (simp_all add: insert_vars_trs_def ac_simps vars_rule_def)

lemma set_insert_funs_trs [simp]:
  "set (insert_funs_trs trs fs) = (\<Union>r \<in> set trs. funs_rule r) \<union> set fs"
  by (induct trs arbitrary: fs) (simp_all add: insert_funs_trs_def ac_simps funs_rule_def)

lemma set_insert_funas_trs [simp]:
  "set (insert_funas_trs trs fs) = (\<Union>r \<in> set trs. funas_rule r) \<union> set fs"
  by (induct trs arbitrary: fs) (simp_all add: insert_funas_trs_def ac_simps funas_rule_def)

lemma set_insert_roots_trs [simp]:
  "set (insert_roots_trs trs fs) = (\<Union>r \<in> set trs. roots_rule r) \<union> set fs"
  by (induct trs arbitrary: fs) (simp_all add: insert_roots_trs_def ac_simps roots_rule_def)

lemma set_insert_funas_args_trs [simp]:
  "set (insert_funas_args_trs trs fs) = (\<Union>r \<in> set trs. funas_args_rule r) \<union> set fs"
  by (induct trs arbitrary: fs)
     (simp_all add: insert_funas_args_trs_def ac_simps funas_args_rule_def)

abbreviation "vars_trs_impl trs \<equiv> insert_vars_trs trs []"
abbreviation "funs_trs_impl trs \<equiv> insert_funs_trs trs []"
abbreviation "funas_trs_impl trs \<equiv> insert_funas_trs trs []"
abbreviation "roots_trs_impl trs \<equiv> insert_roots_trs trs []"
abbreviation "funas_args_trs_impl trs \<equiv> insert_funas_args_trs trs []"

definition defined_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list"
where
  "defined_list R = [the (root l). (l, r) \<leftarrow> R, is_Fun l]"

lemma set_defined_list [simp]:
  "set (defined_list R) = {fn. defined (set R) fn}"
by (force simp: defined_list_def defined_def elim!: root_Some)

definition check_left_linear_trs :: "('f :: show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_left_linear_trs trs =
     check_all (\<lambda>r. linear_term (fst r)) trs
       <+? (\<lambda> _. shows trs \<circ> shows_nl \<circ> shows ''is not left-linear'' \<circ> shows_nl)"

lemma check_left_linear_trs [simp]:
  "isOK (check_left_linear_trs R) = left_linear_trs (set R)"
  unfolding check_left_linear_trs_def left_linear_trs_def
  by auto

definition check_varcond_subset :: "('f :: show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_varcond_subset R =
    check_allm (\<lambda>rule. 
      check_subseteq (vars_term_impl (snd rule)) (vars_term_impl (fst rule))
      <+? (\<lambda>x. shows ''free variable '' \<circ> shows x
        \<circ> shows '' in right-hand side of rule '' \<circ> shows_rule rule \<circ> shows_nl)
    ) R"

lemma check_varcond_subset [simp]:
  "isOK (check_varcond_subset R) = (\<forall> (l, r) \<in> set R. vars_term r \<subseteq> vars_term l)"
  unfolding check_varcond_subset_def by force+

definition "check_varcond_no_Var_lhs =
  check_allm (\<lambda>rule.
    check (is_Fun (fst rule))
      (shows ''variable left-hand side in rule '' \<circ> shows_rule rule \<circ> shows_nl))"

lemma check_varcond_no_Var_lhs [simp]:
  "isOK (check_varcond_no_Var_lhs R) \<longleftrightarrow> (\<forall>(l, r) \<in> set R. is_Fun l)"
by (auto simp: check_varcond_no_Var_lhs_def)

definition check_wf_trs :: "('f :: show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_wf_trs R = do {
    check_varcond_no_Var_lhs R;
    check_varcond_subset R
  } <+? (\<lambda>e. shows ''the TRS is not well-formed'' \<circ> shows_nl \<circ> e)"

lemma check_wf_trs_conf [simp]:
  "isOK (check_wf_trs R) = wf_trs (set R)"
by (force simp: check_wf_trs_def wf_trs_def)

definition check_not_wf_trs :: "('f :: show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_not_wf_trs R =
    check (\<not> isOK (check_wf_trs R)) (shows ''The TRS is well formed'' \<circ> shows_nl)"

lemma check_not_wf_trs:
  assumes "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed

lemma [code]:
  "instance_rule lr st \<longleftrightarrow> match_list (\<lambda> _. fst lr) [(fst st, fst lr), (snd st, snd lr)] \<noteq> None"
  (is "?l = (match_list ?d ?list \<noteq> None)")
proof
  assume ?l
  then obtain \<sigma> where "fst lr = fst st \<cdot> \<sigma>"
    and "snd lr = snd st \<cdot> \<sigma>" by (auto simp: instance_rule_def)
  then have "\<And>l t. (l, t) \<in> set ?list \<Longrightarrow> l \<cdot> \<sigma> = t" by (auto)
  then have "matchers (set ?list) \<noteq> {}" by (auto simp: matchers_def)
  with match_list_complete
    show "match_list ?d ?list \<noteq> None" by blast
next
  assume "match_list ?d ?list \<noteq> None"
  then obtain \<tau> where "match_list ?d ?list = Some \<tau>" by auto
  from match_list_sound [OF this]
    have "fst lr = fst st \<cdot> \<tau>" and "snd lr = snd st \<cdot> \<tau>" by auto
  then show ?l by (auto simp: instance_rule_def)
qed

definition
  check_CS_subseteq :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rule check"
where
  "check_CS_subseteq R S \<equiv> check_allm (\<lambda> (l,r). check (Bex (set S) (instance_rule (l,r))) (l,r)) R"

lemma check_CS_subseteq [simp]:
  "isOK (check_CS_subseteq R S) \<longleftrightarrow> subst.closure (set R) \<subseteq> subst.closure (set S)" (is "?l = ?r")
proof
  assume ?l
  show ?r
  proof
    fix x y
    assume "(x,y) \<in> subst.closure (set R)"    
    thus "(x,y) \<in> subst.closure (set S)"
    proof (induct)
      case (subst s t \<sigma>)
      with \<open>?l\<close>[unfolded check_CS_subseteq_def]
      obtain l r \<delta> where lr: "(l,r) \<in> set S" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>"
        by (auto simp add: instance_rule_def)
      show ?case unfolding s t
        using subst.closure.intros[OF lr, of "\<delta> \<circ>\<^sub>s \<sigma>"]
        by auto
    qed
  qed
next
  assume ?r
  show ?l
    unfolding check_CS_subseteq_def
  proof (simp, intro ballI)
    fix lr
    assume mem: "lr \<in> set R"
    obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
    with mem have "(l,r) \<in> subst.closure (set R)" using subst.subset_closure by auto
    with \<open>?r\<close> have "(l,r) \<in> subst.closure (set S)" by auto
    thus "Bex (set S) (instance_rule lr)" unfolding lr
    proof (induct)
      case (subst s t \<sigma>)
      thus ?case unfolding instance_rule_def by force
    qed
  qed
qed


type_synonym ('f,'v)rule_map = "(('f \<times> nat) \<Rightarrow> ('f,'v)rules)option"


fun computeRuleMapH :: "('f,'v)rules \<Rightarrow> (('f \<times> nat) \<times> ('f,'v)rules)list option"
where "computeRuleMapH [] = Some []"
    | "computeRuleMapH ((Fun f ts,r) # rules) = (let n = length ts in case computeRuleMapH rules of None \<Rightarrow> None | Some rm \<Rightarrow> 
            (case List.extract (\<lambda> (fa,rls). fa = (f,n)) rm of 
                None \<Rightarrow> Some (((f,n), [(Fun f ts,r)]) # rm)
              | Some (bef,(fa,rls),aft) \<Rightarrow>  Some  ((fa,(Fun f ts,r) # rls) # bef @ aft)))"
    | "computeRuleMapH ((Var _, _) # rules) = None"

definition computeRuleMap :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rule_map" where
  "computeRuleMap rls \<equiv>
    (case computeRuleMapH rls of
      None \<Rightarrow> None
    | Some rm \<Rightarrow> Some (\<lambda>f.
      (case map_of rm f of
        None \<Rightarrow> []
      | Some rls \<Rightarrow> rls)))"

lemma computeRuleMapHSound2: "(computeRuleMapH R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
proof (induct R, simp)
  case (Cons rule rules)
  show ?case 
  proof (cases rule)
    case (Pair l r)
    show ?thesis 
    proof (cases l, simp add: Pair Cons)
      case (Fun f ts)
      show ?thesis
        using Pair Cons
      proof (cases "computeRuleMapH rules", simp add: Pair Fun Let_def)
        case (Some rm) note oSome = this
        let ?e = "List.extract (\<lambda> (fa,rls). fa = (f,length ts)) rm"
        from Some Fun Pair Cons show ?thesis 
        proof (cases ?e, simp)
          case (Some res)
          then obtain bef aft g rls where "?e = Some (bef, (g,rls), aft)" by (cases res, force) 
          with extract_SomeE[OF this] have rm: "rm = bef @ ((f, length ts),rls) # aft" and e: "?e = Some (bef, ((f,length ts),rls), aft)"
            by auto
          show ?thesis using Cons
            by (simp add: Pair Fun Let_def oSome e)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound2: "(computeRuleMap R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
unfolding computeRuleMap_def
by (simp only: computeRuleMapHSound2[symmetric], cases "computeRuleMapH R", auto)


lemma computeRuleMapHSound: assumes "computeRuleMapH R = Some rm"
  shows "(\<lambda> (f,rls). (f,set rls)) ` set rm = {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)} \<noteq> {}} \<and> distinct_eq (\<lambda> (f,rls) (g,rls'). f = g) rm"
using assms
proof (induct R arbitrary: rm, force)
  case (Cons rule rules)
  let ?setl = "\<lambda> rm. (\<lambda> (f,rls). (f,set rls)) ` set rm"
  let ?setr = "\<lambda> R. {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)} \<noteq> {}}"
  show ?case
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var v)
      with Cons Pair show ?thesis by simp
    next
      case (Fun f ts)
      with Cons Pair show ?thesis 
      proof (cases "computeRuleMapH rules", simp)
        case (Some rrm) note oSome = this
        let ?dis = "distinct_eq (\<lambda> (f,rls) (g,rls'). f = g)"
        from Cons(1)[OF Some] have drrm: "?dis rrm" and srrm: "?setl rrm = ?setr rules" by auto
        show ?thesis
        proof (cases "List.extract (\<lambda> (fa,rls). fa = (f,length ts)) rrm")
          case None
          let ?e = "((f,length ts), [(Fun f ts,r)])"
          let ?e' = "((f,length ts), {(Fun f ts,r)})"
          from None Cons(2) have rm: "rm = ?e # rrm" by (simp add: Fun Pair Some None)
          from None[unfolded extract_None_iff] have rrm: "\<And> g n rl. ((g,n),rl) \<in> set rrm \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by auto
          hence rrm': "\<And> g n rl. ((g,n),rl) \<in> ?setr rules \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by (simp only: srrm[symmetric], auto)
          hence  id: " {(Fun f ts, r)} = {(l, ra). (l = Fun f ts \<and> ra = r \<or> (l, ra) \<in> set rules) \<and> root l = Some (f, length ts)}"  by force
          from rrm have dis: "?dis rm" 
            by (simp add: rm drrm, auto)
          have "?setl rm = insert ?e' (?setl rrm)" by (simp add: rm)
          also have "\<dots> = insert ?e' (?setr rules)" by (simp add: srrm)
          also have "\<dots> = ?setr ( (Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rls
            show "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False
              hence "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr rules)" by auto
              also have "\<dots> = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rls = {(Fun f ts, r)}")
                case True
                with oTrue show ?thesis by (simp add: id, force)
              next
                case False
                show ?thesis using rrm'[of g n rls] True False by (simp add: False True id, auto)
              qed
            qed
          qed
          finally show ?thesis
            by (simp only: dis drrm, simp add: Pair Fun)
        next
          case (Some res)
          obtain bef fg rls aft where "res = (bef,(fg,rls),aft)" by (cases res, force)
          from extract_SomeE[OF Some[simplified this]] Some[simplified this] have rrm: "rrm = bef @ ((f,length ts), rls) # aft" 
            and e: "List.extract (\<lambda> (fa, rls). fa = (f,length ts)) rrm = Some (bef, ((f,length ts),rls),aft)" by auto
          let ?e = "((f,length ts), (Fun f ts,r) # rls)"
          let ?e' = "((f,length ts), insert (Fun f ts,r) (set rls))"
          have "((f,length ts),set rls) \<in> ?setl rrm" unfolding rrm by auto
          hence rls: "set rls = {(l, r) |l r. (l, r) \<in> set rules \<and> root l = Some (f, length ts)}" using Cons(1)[OF oSome] by auto
          obtain ba where ba: "ba = bef @ aft" by auto
          from Cons(2) e ba have rm: "rm = ?e # ba" by (simp add: Fun Pair oSome e)
          from drrm[simplified rrm] have dis: "?dis ba"  unfolding distinct_eq_append ba by auto
          from drrm[simplified rrm] have dis: "?dis rm" unfolding rm distinct_eq_append ba by (auto simp: dis[simplified ba])
          from drrm[simplified rrm distinct_eq_append] 
          have diff: "(\<forall>x\<in>set ba. \<not> (\<lambda>(g, rls). (f,length ts) = g) x)" by (auto simp: ba)
          have "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setl rrm" using rrm ba by auto
          also have "\<dots> = ?setr rules" by (rule srrm)
          finally have id: "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setr rules" .          
          have "?setl rm = insert ?e' (?setl ba)" by (simp add: rm)
          also have "\<dots> = ?setr ((Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rl 
            show "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False 
              hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setl ba)" by auto
              also have "\<dots> = (((g,n),rl) \<in> ?setr rules)" using False by (simp only: id[symmetric], auto)
              also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rl = insert (Fun f ts, r) (set rls)")
                case True
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = True" using oTrue by auto
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" unfolding True rls using oTrue by force
                finally show ?thesis .
              next
                case False
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = False" using diff by (simp add: True, auto)
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" 
                proof (rule ccontr)
                  assume "\<not> ?thesis"
                  with True have "((f,length ts),rl) \<in> ?setr ((Fun f ts,r) # rules)" by simp
                  hence "rl = {(l, ra) |l ra. (l, ra) \<in> set ((Fun f ts, r) # rules) \<and> root l = Some (f, length ts)}" by simp
                  with False rls show False by auto
                qed                              
                finally show ?thesis .
              qed
            qed
          qed
          also have "\<dots> = ?setr (rule # rules)" by (simp add: Pair Fun)
          finally show ?thesis by (simp add: dis)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound:
  assumes "computeRuleMap R = Some rm" 
  shows "(set (rm (f,n))) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
proof (cases "computeRuleMapH R")
  case None
  thus ?thesis using assms unfolding computeRuleMap_def by auto
next
  case (Some rrm)
  note rrm = computeRuleMapHSound[OF this] 
  note rm = assms[unfolded computeRuleMap_def, simplified Some, simplified, symmetric]
  show ?thesis
  proof (cases "map_of rrm (f, n)")
    case (Some rls)
    from map_of_SomeD[OF this] have "((f,n),set rls) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm"
      by auto
    hence "set rls = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some (f, n)}"
      by (simp only: rrm, simp) 
    thus ?thesis by (simp add: rm Some)
  next
    case None
    have id: "{(l, r) |l r. (l, r) \<in> set R \<and> root l = Some (f, n)} = {}" (is "?set = {}")
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then obtain l r where "(l, r) \<in> set R \<and> root l = Some (f, n)" by auto
      with rrm have "(((f,n), ?set)) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm" by auto
      with None[unfolded map_of_eq_None_iff] show False by force
    qed
    thus ?thesis by (simp only: rm None id, auto)
  qed
qed
    
    
  
lemma computeRuleMap_left_vars:
  shows "(computeRuleMap R \<noteq> None) = (\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x)"
proof (cases "computeRuleMap R")
  case None
  from None computeRuleMapSound2 have "\<exists> (l,r) \<in> set R. root l = None" by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this have "(l,r) \<in> set R \<and> \<not> (\<forall> x. fst (l,r) \<noteq> Var x)" by (cases l, auto)
  with None show ?thesis by blast
next
  case (Some rm)
  hence left: "computeRuleMap R \<noteq> None" by auto
  from Some computeRuleMapSound2 have "\<forall> (l,r) \<in> set R. root l \<noteq> None" by force
  hence "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" by auto
  with left show ?thesis by blast
qed

lemma computeRuleMap_defined: fixes R :: "('f,'v)rules"
  assumes "computeRuleMap R = Some rm"
  shows "(rm (f,n) = []) = (\<not> defined (set R) (f,n))"
proof -
  from assms computeRuleMapSound have rm: "\<And>(f::'f) n. set (rm (f,n)) = {(l,r) | l r. (l, r) \<in> set R \<and> root l = Some (f, n)}" by force
  show ?thesis
  proof (cases "rm (f,n)")
    case Nil
    with rm have "\<not> defined (set R) (f,n)" unfolding defined_def by force
    with Nil show ?thesis by blast
  next
    case (Cons lr RR)
    hence left: "rm (f,n) \<noteq> []" by auto
    from Cons rm[where f = f and n = n] have "defined (set R) (f,n)" unfolding defined_def by (cases lr, force)
    with left show ?thesis by blast
  qed
qed



lemma computeRuleMap_None_not_SN:
  assumes "computeRuleMap R = None"
  shows "\<not> SN_on (rstep (set R)) {t}"
proof -
  from assms computeRuleMap_left_vars[of R] obtain x r where "(Var x,r) \<in> set R" by auto
  from left_var_imp_not_SN[OF this] show ?thesis .
qed


definition reverse_rules :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules" where
  "reverse_rules rs \<equiv> map prod.swap rs"

lemma reverse_rules[simp]: "set (reverse_rules R) = (set R)^-1" unfolding reverse_rules_def by force

definition
  map_funs_rules_wa :: "('f \<times> nat \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('g, 'v) rules"
where
  "map_funs_rules_wa fg R = map (\<lambda>(l, r). (map_funs_term_wa fg l, map_funs_term_wa fg r)) R"

lemma map_funs_rules_wa: "set (map_funs_rules_wa fg R) = map_funs_trs_wa fg (set R)"
  unfolding map_funs_rules_wa_def map_funs_trs_wa_def by auto

lemma wf_rule [code]:
  "wf_rule r \<longleftrightarrow>
    is_Fun (fst r) \<and> (\<forall> x \<in> set (vars_term_impl (snd r)). x \<in> set (vars_term_impl (fst r)))"
  unfolding wf_rule_def by auto

definition wf_rules_impl :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules"
where
  "wf_rules_impl R = filter wf_rule R"

lemma wf_rules_impl [simp]:
  "set (wf_rules_impl R) = wf_rules (set R)"
  unfolding wf_rules_impl_def wf_rules_def by auto

fun check_wf_reltrs :: "('f::show, 'v::show) rules \<times> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_wf_reltrs (R, S) = (do {
     check_wf_trs R;
     if R = [] then succeed
     else check_varcond_subset S
   })"

lemma check_wf_reltrs[simp]:
  "isOK (check_wf_reltrs (R, S)) = wf_reltrs (set R) (set S)"
  by (cases R) auto

declare check_wf_reltrs.simps[simp del]

definition
  check_linear_trs :: "('f::show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_linear_trs R \<equiv>
     check_all (\<lambda> (l,r). (linear_term l) \<and> (linear_term r)) R
       <+? (\<lambda> _. shows R \<circ> shows_nl \<circ> shows ''is not linear'' \<circ> shows_nl)"

lemma check_linear_trs [simp]:
  "isOK (check_linear_trs R) \<longleftrightarrow> linear_trs (set R)"
  unfolding check_linear_trs_def linear_trs_def by auto

definition "non_collapsing_impl R = list_all (is_Fun o snd) R"

lemma [simp]: "non_collapsing_impl R = non_collapsing (set R)"
  unfolding non_collapsing_impl_def non_collapsing_def list_all_iff by auto


type_synonym ('f, 'v) term_map = "'f \<times> nat \<Rightarrow> ('f, 'v) term list"

definition term_map :: "('f::key, 'v) term list \<Rightarrow> ('f, 'v) term_map" where
  "term_map ts = fun_of_map (rm.\<alpha> (elem_list_to_rm (the \<circ> root) ts)) []"

definition
  is_NF_main :: "bool \<Rightarrow> bool \<Rightarrow> ('f::key, 'v) term_map \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_NF_main var_cond R_empty m = (if var_cond then (\<lambda>_. False) 
     else if R_empty then (\<lambda>_. True)
     else (\<lambda>t. \<forall>u\<in>set (supteq_list t).
       if is_Fun u then
         \<forall>l\<in>set (m (the (root u))). \<not> matches u l
       else True))"

lemma neq_root_no_match:
  assumes "is_Fun l" and "the (root l) \<noteq> the (root t)"
  shows "\<not> matches t l"
  using assms by (cases t) (force iff: matches_iff)+

lemma all_not_conv: "(\<forall>x \<in> A. \<not> P x) = (\<not> (\<exists>x \<in> A. P x))" by auto

lemma efficient_supteq_list_do_not_match:
  assumes "\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
  shows
    "(\<forall>l\<in>set ls. \<forall>u\<in>set (supteq_list t). \<not> matches u l) \<longleftrightarrow>
     (\<forall>u\<in>set (supteq_list t). \<forall>l\<in>set (term_map ls (the (root u))).
       \<not> matches u l)"
    (is "?lhs \<longleftrightarrow> ?rhs" is "_ \<longleftrightarrow> (\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l)")
proof (intro iffI ballI)
  fix u l assume "\<forall>l\<in>set ls. \<forall>u\<in>set ?subs. \<not> matches u l" and "u \<in> set ?subs"
    and "l \<in> set (?ls u)"
  thus "\<not> matches u l"
    using elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ls "the (root u)"]
    by (auto simp: o_def term_map_def rm_set_lookup_def)
next
  fix l u assume 1: "\<forall>u\<in>set ?subs. \<forall>l\<in>set (?ls u). \<not> matches u l"
    and "l \<in> set ls" and "u \<in> set ?subs"
  with assms have "the (root l) \<noteq> the (root u) \<longrightarrow> \<not> matches u l"
    and "\<forall>l\<in>set (?ls u). \<not> matches u l " by simp+
  with elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ls "the (root u)"]
    and \<open>l \<in> set ls\<close>
    show "\<not> matches u l" by (auto simp: o_def term_map_def rm_set_lookup_def)
qed

lemma supteq_list_ex:
  "(\<exists>u\<in>set (supteq_list l). \<exists>\<sigma>. t \<cdot> \<sigma> = u) \<longleftrightarrow> (\<exists>\<sigma>. l \<unrhd> t \<cdot> \<sigma>)"
  unfolding supteq_list by auto

definition "is_NF_trs R = is_NF_main (\<exists>r\<in>set R. is_Var (fst r)) (R = []) (term_map (map fst R))"
definition "is_NF_terms Q = is_NF_main (\<exists>q\<in>set Q. is_Var q) (Q = []) (term_map Q)"

lemma is_NF_main_NF_trs_conv:
  "is_NF_main (\<exists>r\<in>set R. is_Var (fst r)) (R = []) (term_map (map fst R)) t \<longleftrightarrow>
    t \<in> NF_trs (set R)"
  (is "is_NF_main ?var ?R ?map t \<longleftrightarrow> _")
proof (intro iffI allI)
  assume is_NF_main: "is_NF_main ?var ?R ?map t"
  show "t \<in> NF_trs (set R)"
  proof (cases "\<exists>r\<in>set R. is_Var (fst r)")
    case True with is_NF_main[unfolded is_NF_main_def] show ?thesis by simp
  next
    case False
    let ?ts = "map fst R"
    from False have allfun: "\<forall>s\<in>set ?ts. is_Fun s" by simp
    with neq_root_no_match
    have no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
        \<longrightarrow> \<not> matches u s" by blast
    note is_NF_main = is_NF_main[unfolded is_NF_main_def if_not_P[OF False]]
    show ?thesis
    proof (cases "R = []")
      case False
      hence False: "(R = []) = False" by simp
      have "\<forall>u\<in>set (supteq_list t). \<forall>l\<in>set (term_map ?ts (the (root u))). \<not> matches u l"
      proof
        fix u
        assume mem: "u \<in> set (supteq_list t)"
        show "\<forall>l\<in>set (term_map ?ts (the (root u))). \<not> matches u l"
        proof (cases u)
          case (Var x)
          show ?thesis
          proof
            fix l
            assume "l \<in> set (term_map ?ts (the (root u))) "
            with elem_list_to_rm.rm_set_lookup[of "the \<circ> root" ?ts "the (root u)"]
            have "l \<in> set ?ts" by (auto simp: o_def term_map_def rm_set_lookup_def)
            hence "is_Fun l" using allfun by auto
            hence "(\<forall>\<sigma>. l \<cdot> \<sigma> \<noteq> u)" using Var by auto
            thus "\<not> matches u l" using matches_iff by blast
          qed
        next
          case (Fun f us)
          with mem is_NF_main[unfolded False] show ?thesis by auto
        qed
      qed
      thus ?thesis
        unfolding efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches_iff
        unfolding supteq_list_ex by auto
    qed auto
  qed
next
  assume NF_trs: "t \<in> NF_trs (set R)"
  show "is_NF_main (\<exists>r\<in>set R. is_Var (fst r)) (R = []) (term_map (map fst R)) t"
  proof (cases "\<exists>r\<in>set R. is_Var (fst r)")
    case True
    then obtain l where "l \<in> lhss (set R)" and "is_Var l" by auto
    from lhs_var_not_NF[OF this] and NF_trs show ?thesis by simp
  next
    case False note oFalse = this
    let ?ts = "map fst R"
    from False have "\<forall>s\<in>set ?ts. is_Fun s" by auto
    with neq_root_no_match
    have
      no_match: "\<forall>s\<in>set ?ts. \<forall>u\<in>set (supteq_list t). the (root s) \<noteq> the (root u)
      \<longrightarrow> \<not> matches u s" by blast
    show ?thesis
    proof (cases "R = []")
      case True thus ?thesis unfolding is_NF_main_def by auto
    next
      case False
      hence False: "(R = []) = False" by simp
      from NF_trs
      show ?thesis
        unfolding is_NF_main_def False if_False if_not_P[OF oFalse]
        using efficient_supteq_list_do_not_match[OF no_match, symmetric]
        unfolding all_not_conv matches_iff
        unfolding supteq_list_ex set_map by fastforce
    qed
  qed
qed

lemma is_NF_trs [simp]:
  "is_NF_trs R = (\<lambda> t. t \<in> NF_trs (set R))"
  by (rule ext, unfold is_NF_trs_def is_NF_main_NF_trs_conv, simp)

lemma is_NF_terms [simp]:
  "is_NF_terms Q = (\<lambda> t. t \<in> NF_terms (set Q))"
proof (rule ext)
  fix t
  let ?Q = "map (\<lambda>t. (t, t)) Q"
  have 1: "(\<exists>t\<in>set Q. is_Var t) = (\<exists>t\<in>set ?Q. is_Var (fst t))"
    by (induct Q) (auto simp: o_def)
  have 2: "map fst ?Q = Q" by (induct Q) simp_all
  have 3: "term_map Q = term_map (map fst ?Q)" unfolding 2 ..
  have 4: "set ?Q = Id_on (set Q)" by (induct Q) (auto simp: o_def)
  from is_NF_main_NF_trs_conv[of "?Q" t]
  show "is_NF_terms Q t = (t \<in> NF_terms (set Q))"
    unfolding is_NF_terms_def 1 3 4 unfolding 2 by auto
qed

abbreviation "terms_of_rules rs \<equiv> map fst rs @ map snd rs"
  
definition "match_rules rs\<^sub>1 rs\<^sub>2 = do {
  ts \<leftarrow> zip_option (terms_of_rules rs\<^sub>2) (terms_of_rules rs\<^sub>1);
  match_list Var ts
}"

abbreviation "term_of_rules rs \<equiv> Fun (SOME f. True) (terms_of_rules rs)"

lemma match_rules_alt_def:
  "match_rules rs\<^sub>1 rs\<^sub>2 = match (term_of_rules rs\<^sub>1) (term_of_rules rs\<^sub>2)"
  by (cases "zip_option (terms_of_rules rs\<^sub>1) (terms_of_rules rs\<^sub>2)")
    (auto simp: match_rules_def match_def match_list_def decompose_def
          split: option.splits Option.bind_splits)

lemma rules_enc_aux:
  fixes xs ys :: "('f, 'v :: infinite) rules"
  assumes "map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs = map fst ys @ map snd ys"
  shows "map (op \<bullet> \<pi>) xs = ys"
proof -
  have "length xs = length ys"
  proof (rule ccontr)
    assume "length xs \<noteq> length ys"
    moreover
    have "length (map (op \<bullet> \<pi> \<circ> fst) xs @ map (op \<bullet> \<pi> \<circ> snd) xs) = length (map fst ys @ map snd ys)"
      using assms by simp
    ultimately show False by simp
  qed
  then show ?thesis
    using assms by (induct xs ys rule: list_induct2; auto simp: eqvt)
qed

lemma perm_term_of_rules_eq_conv:
  "\<pi> \<bullet> term_of_rules rs\<^sub>1 = term_of_rules rs\<^sub>2 \<longleftrightarrow> \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
  by (auto simp: eqvt rules_enc_aux)

lemma match_rules_imp_variants:
  fixes rs\<^sub>1 rs\<^sub>2 :: "('f, 'v :: infinite) rules"
  assumes "match_rules rs\<^sub>2 rs\<^sub>1 = Some \<sigma>\<^sub>1" and "match_rules rs\<^sub>1 rs\<^sub>2 = Some \<sigma>\<^sub>2"
  shows "\<exists>\<pi>::'v perm. \<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2"
proof -
  have "term_of_rules rs\<^sub>2 \<cdot> \<sigma>\<^sub>2 = term_of_rules rs\<^sub>1"
    and "term_of_rules rs\<^sub>1 \<cdot> \<sigma>\<^sub>1 = term_of_rules rs\<^sub>2"
    using assms by (auto simp: match_rules_alt_def dest: match_sound)
  then obtain \<pi> :: "'v perm" where "\<pi> \<bullet> rs\<^sub>2 = rs\<^sub>1"
    using term_variants_iff [of "term_of_rules rs\<^sub>2" "term_of_rules rs\<^sub>1"]
    unfolding perm_term_of_rules_eq_conv by auto
  then have "-\<pi> \<bullet> rs\<^sub>1 = rs\<^sub>2" by (auto simp del: rules_pt.permute_list_def)
  then show ?thesis ..
qed

definition
  "check_variants_rule r r' = do {
    check (match_rules [r] [r'] \<noteq> None \<and> match_rules [r'] [r] \<noteq> None)
      (shows_rule r \<circ> shows '' and '' \<circ> shows_rule r' \<circ> shows '' are not variants of each other'' \<circ> shows_nl)
  }"

lemma check_variants_rule [elim!]:
  assumes "isOK (check_variants_rule r r')"
  obtains p where "p \<bullet> r = r'"
  using assms
  by (cases r; cases r') (auto simp: check_variants_rule_def eqvt dest: match_rules_imp_variants)

definition
  "check_variant_in_trs R r =
    check_exm (check_variants_rule r) R (shows_sep id id)
    <+? (\<lambda>_. shows_rule r \<circ> shows '' is not a variant of any rule in '' \<circ> shows_trs R)"

lemma check_variant_in_trs [elim!]:
  assumes "isOK (check_variant_in_trs R r)"
  obtains p where "p \<bullet> r \<in> set R"
  using assms by (cases r) (force simp: check_variant_in_trs_def eqvt)

definition "check_variants_trs R R' = check_allm (check_variant_in_trs R') R"

lemma check_variants_trs [dest!]:
  assumes "isOK (check_variants_trs R R')"
  shows "\<forall>r \<in> set R. \<exists>p. p \<bullet> r \<in> set R'"
  using assms by (auto simp: check_variants_trs_def)

definition "check_litsim_trs R R' = do {
  check_variants_trs R R';
  check_variants_trs R' R
} <+? (\<lambda>e. shows_trs R \<circ> shows_nl \<circ> shows ''is not literally similar to '' \<circ> shows_trs R' \<circ> e)"

lemma check_litsim_trs [dest!]:
  assumes "isOK (check_litsim_trs R R')"
  shows "set R \<doteq> set R'"
  using assms by (auto simp: check_litsim_trs_def subsumable_trs.litsim_def)

end
